/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

 
//                          zi version: 1.0


//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: 

var variable = '1';  // Variable selector: '1' = Intersect angle; '2' = Big angle '3' = Small lines; '4' = Big lines; '5' = Big length 

var iter = 1;   // Number of iterations for each variable value

var back_clr = '#fff';  // background canvas color during task

var pen_clr = '#000';  //  font color (base font) during the task 

var pen_clr_c = '#006400'; //  font color for a demo mode (show ideal line/solution)

var pen_width = 1;   // Pen width for all lines

var is_circle = false;  // draw a outer circle around stimulus 
  
var is_back = true;    // if (is_back==true) => user can go back to previous task/item.  

var prg_type = true;   //  Progress Bar is visible (true) or invisible (false)

var demo = false;     // if (demo == true) => Show task solutions 

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other global constants and variables:

var item_cnt;       // number of items in the task. range [0..item_cnt-1]

var item;   // current item in task

var is_task = false;    // task in progress or not for pageResize() event. It is true for all events between end of startTask() and theEnd() events

var dirty = false;  // if  (dirty==true) => user changed any options in form 

var is_storage= false; // local storage is enabled on user browser or not

var out_range;        // feedback message to user on footer: Out of range. Press the opposite arrow key 
  
//  global objects and arrays declarations..
 
var stim = {};   // stimulus object. see: stimGen()  

var prop = {};  // stimulus properties objects. see setPar() => { 'angle': angle[i], 'gap': gap[i], 'rotation': rot[i], 'length': size[i] }; 

var dv = [];  // dependent variable for each item in range (error-pixels and percents). dimensions: [0 .. item_cnt-1] * [0 == error(px), 1 == percent]. 

var mis = [];  // missing values for each item in range: [0..item_cnt-1]
	


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */



// application init [check storage, set default values, show start panel..] 
function init() 
{ 
  setDefault();	        
	is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }	
	document.onkeydown = onKeyDown;
	document.onkeyup = onKeyUp;
	setIns();
	showStart();  
}

 
// set default values in all <textarea> objects => variables
function setDefault()
{ 
	document.getElementById('i_angle').value= '20\n15\n25';
	document.getElementById('g_angle').value= '45\n0\n90';
	document.getElementById('small_cnt').value= '20\n10\n5';
	document.getElementById('big_cnt').value= '8\n4\n10';
	document.getElementById('big_size').value= '100\n50\n25';
	//   \r\n   or    \n  ?
}


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
		return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 (probably redundant?)
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// raised when page is resized. see html: <body onresize="pageResize()"
function pageResize()  
{ 
	if ( is_task )  { 
    theEnd(true);	  // resize error			
	}
}	


// show start panel 
function showStart()  
{     
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (is_storage && localStorage.getItem('zi-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('zi-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';  // xxx
		}
	}
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add('center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';          
  document.getElementById('hide_option').style.display = 'none';    
  document.getElementById('instructions').style.display= 'block';
	refMsg(); 
  setIns();	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		
  document.getElementById('ins_back').style.display= is_back ? 'none' : 'block';
	document.getElementById('ins_demo').style.display= demo ? 'block' : 'none';			
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty = true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 'variable': 
			msg.textContent= e.options[e.selectedIndex].getAttribute('data-help');	
		  for (let i= 1; i<=5; i++)  {
				document.getElementById('fs_' + i).style.backgroundColor= ( e.value == i ) ? '#e5e5e5': '#fff'; 
			} 
		  break;
			
		case 'iter':
			if (e.validity.valid)  {
				msg.textContent= 'Total number of items = (number of iterations) * (number of entered valid values ​​for the selected variable)'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 
				
    case 'back_clr': 		// background canvas color during task
		case 'pen_clr':  		// font color (base font) during the task 
		case 'pen_clr_c':		// font color for a demo mode       
			msg.textContent= ( validColor(e.value) ) ? '' : 'The selected color is invalid.'; 
			break; 		
			
    case 'pen_width':
			if (e.validity.valid)  {
				msg.textContent= ''; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 
			
		case 'is_circle':
		case 'is_back':		
			msg.textContent= e.checked ? 'Yes.' : 'No.';      
			break;  	
		
    case 'prg_type':                                                                                                                                    
      msg.textContent= e.checked ? 'Yes. It will be displayed how many items are left to the end.' : 'No.';      
      break;    

		case 'demo': 
			msg.textContent= e.checked ? 'Yes. The solution of each task will be displayed (default: green line). The error size will be displayed in the middle of the footer.' : 'No';  
      break; 
		
		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty === true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity in most of options (except all textarea elements)
function optionsValid()  
{ 
  let res = document.getElementById('iter').validity.valid && document.getElementById('pen_width').validity.valid  &&  validColor(document.getElementById('back_clr').value) && validColor(document.getElementById('pen_clr').value) && validColor(document.getElementById('pen_clr_c').value);
  return res;              
}


// start task (research) 
function startTask(e)   
{      			
  initLayout(); 	
   
  setPar();
 
	stimGen();
	
	is_task= true;
	
	drawStim(true);    // draw first item (stimulus)
}


// init DOM/html elements (positions,visibility,..)
function initLayout()
{	
	//document.getElementById('result_wrapper').style.display= 'none';
  //document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
	document.getElementById('task_section').style.display= 'block'; 	  	
	document.getElementById('progress_holder').style.display = document.getElementById('prg_type').checked ? 'block' : 'none';
  document.getElementById('nav_items').textContent= '';	
	
	setCanvas();
}  


// set canvas absolute positions on screen
function setCanvas()
{
	// canvas must be after set footer content => progress_holder visible or not
	let c= document.getElementById('can');
	c.style.display= 'block';  
	c.width= innerWidth;  
	c.style.width= c.width + 'px';	     
	c.height= innerHeight - document.getElementById('footer').offsetHeight; // document.getElementById('footer').clientHeight == document.getElementById('footer').offsetHeight  
	c.style.height= c.height + 'px';  // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 			
  
  c.getContext('2d').lineWidth= pen_width; 		// set Pen width for all lines
}


// validates and filter values entered in all textarea elements. set default values
function validate()
{
	// Variable selector: '1' = Intersect angle; '2' = Big angle '3' = Small lines; '4' = Big lines; '5' = Big length
	
	let i_angle= [], g_angle= [], small_cnt = [], big_cnt= [], big_size= [], v;
	
	// 1. Intersect angle between big and small lines (id="i_angle"). range= [5-90]. default = 20.
	let lines = document.getElementById('i_angle').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		lines[i]= trim(lines[i]);  // code here using lines[i] which will give you each line
		if ( isInt(lines[i]) )  { 
		  v = parseInt(lines[i], 10);
			if ( v >= 5  &&  v <= 90 )  {
				i_angle.push(v);
			}
		}
	}
	if (i_angle.length == 0)  {
		i_angle.push(20);
	}	 	
	document.getElementById('i_angle').value= variable == '1' ? i_angle.join('\n') : i_angle[0];
	
	// 2. Angle of big lines (id="g_angle"). range= [0..90]. Units: degrees. default: 45 
	lines = document.getElementById('g_angle').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		lines[i]= trim(lines[i]);
		if ( isInt( lines[i]) )  { 
		  v = parseInt(lines[i], 10);
			if ( v >= 0  &&  v <= 90 )  {
				g_angle.push(v);
			}
		}
	}
	if (g_angle.length == 0)  {
		g_angle.push(45);
	}	
	document.getElementById('g_angle').value= variable == '2' ? g_angle.join('\n') : g_angle[0];

	// 3. Number of small lines (id="small_cnt"). range= [1..50]. default: 10
	lines = document.getElementById('small_cnt').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		lines[i]= trim(lines[i]);
		if ( isInt( lines[i]) )  {  
		  v = parseInt(lines[i], 10);
			if ( v >= 1  &&  v <= 50 )  {
				small_cnt.push(v);
			}
		}
	}
	if (small_cnt.length == 0)  {
		small_cnt.push(10);
	}	
	document.getElementById('small_cnt').value= variable == '3' ? small_cnt.join('\n') : small_cnt[0];
	
	// 4. Number of big lines (id="big_cnt"). range= [2..50]. default: 2
	lines = document.getElementById('big_cnt').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		lines[i]= trim(lines[i]);
		if ( isInt( lines[i]) )  {  
		  v = parseInt(lines[i], 10);
			if ( v % 2 == 0 && v >= 2  &&  v <= 50 )  {
				big_cnt.push(v);
			}
		}
	}
	if (big_cnt.length == 0)  {
		big_cnt.push(2);
	}	
	document.getElementById('big_cnt').value= variable == '4' ? big_cnt.join('\n') : big_cnt[0];
	
	// 5. Size of big lines (id="big_size"). range= [10..100]. default: 100
	lines = document.getElementById('big_size').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		lines[i]= trim(lines[i]);
		if ( isInt( lines[i]) )  {  
		  v = parseInt(lines[i], 10);
			if ( v >= 10  &&  v <= 100 )  {
				big_size.push(v);
			}
		}
	}
	if (big_size.length == 0)  {
		big_size.push(100);
	}	
	document.getElementById('big_size').value= (variable == '5') ? big_size.join('\n') : big_size[0];
	
}
	
	
/* set internal application research parameters (global vars)
   set stimulus properties (variables) for stimulus generator  */
function setPar()  
{  
	item = 0; 

	stim = {};	
	prop = {};
  dv = [];
	mis = []; 	 
	
	validate();   // filter and validate variable values (all textarea elements)
	
	// prepare stimulus properties for stimulus generator	
	// Variable selector: '1' = Intersect angle; '2' = Big angle '3' = Small lines; '4' = Big lines; '5' = Big length
	
	let i_angle= [], g_angle= [], small_cnt = [], big_cnt= [], big_size= [], v;
	
	// 1. Intersect angle between big and small lines (id="i_angle")
	let lines = document.getElementById('i_angle').value.split('\n');
	for (let i = 0; i < lines.length; i++)  {
		v = parseInt(lines[i], 10);
		for (let j= 1; j<= iter; j++) {
			i_angle.push(v);
		}
	}	
	if  (variable == '1' && i_angle.length > 1) {
		i_angle= shuffle(i_angle);
	}
	
	// '2' = Angle of big lines (id="g_angle")
	lines = document.getElementById('g_angle').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		v = parseInt(lines[i], 10);
		for (let j= 1; j<= iter; j++) {
			g_angle.push(v);
		}
	}
	if (variable == '2' && g_angle.length > 1) {
		g_angle= shuffle(g_angle);
	}
	
	// 3. Number of small lines (id="small_cnt")
	lines = document.getElementById('small_cnt').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		v = parseInt(lines[i], 10);
		for (let j= 1; j<= iter; j++) {
			small_cnt.push(v);
		}
	}
	if  (variable ==  '3' && small_cnt.length > 1) {
		small_cnt= shuffle(small_cnt);
	}
	
	// 4. Number of big lines (id="big_cnt")
	lines = document.getElementById('big_cnt').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		v = parseInt(lines[i], 10);
		for (let j= 1; j<= iter; j++) {
			big_cnt.push(v);
		}
	}
	if  (variable == '4' && big_cnt.length > 1) {
		big_cnt= shuffle(big_cnt);
	}
	
	// 5. Size of big lines (id="big_size")
	lines = document.getElementById('big_size').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		v = parseInt(lines[i], 10);
		for (let j= 1; j<= iter; j++) {
			big_size.push(v);
		}
	}
	if  (variable == '5' && big_size.length > 1) {
		big_size= shuffle(big_size);
	}
	
	
  switch (variable)
	{
		case '1': 
			item_cnt= i_angle.length;
			for (let i=0; i < item_cnt; i++)  {
				prop[i] = { 'i_angle': i_angle[i], 'g_angle': g_angle[0], 'small_cnt': small_cnt[0], 'big_cnt': big_cnt[0], 'big_size': big_size[0] }; 
			}
			break;
		case '2': 
			item_cnt= g_angle.length;
			for (let i=0; i < item_cnt; i++)  {
				prop[i] = { 'i_angle': i_angle[0], 'g_angle': g_angle[i], 'small_cnt': small_cnt[0], 'big_cnt': big_cnt[0], 'big_size': big_size[0] }; 
			}
			break;
		case '3': 
			item_cnt= small_cnt.length;
			for (let i=0; i < item_cnt; i++)  {
				prop[i] = { 'i_angle': i_angle[0], 'g_angle': g_angle[0], 'small_cnt': small_cnt[i], 'big_cnt': big_cnt[0], 'big_size': big_size[0] }; 
			}	
			break;		
		case '4': 
			item_cnt= big_cnt.length;
			for (let i=0; i < item_cnt; i++)  {
				prop[i] = { 'i_angle': i_angle[0], 'g_angle': g_angle[0], 'small_cnt': small_cnt[0], 'big_cnt': big_cnt[i], 'big_size': big_size[0] }; 
			}		
			break;
		case '5': 
			item_cnt= big_size.length;
			for (let i=0; i < item_cnt; i++)  {
				prop[i] = { 'i_angle': i_angle[0], 'g_angle': g_angle[0], 'small_cnt': small_cnt[0], 'big_cnt': big_cnt[0], 'big_size': big_size[i] }; 
			}		
			break;				
	}
  
	// after setting item_cnt
	for (let i= 0; i< item_cnt; i++) { 
    mis[i]= true;	
		dv[i]= [null,null];	 
  }
}


// generate all stimuli for each item in range 0..item_cnt-1	
function stimGen()	
{			 	
	let c= document.getElementById('can');	
	for (let i = 0; i < item_cnt; i++)  {	
	  stim[i] = {};
		setStim(i);	
	}  
}


/*  set stimulus (item) coord
    n = item number. range 0..item_cnt-1  */	
function setStim(n)
{  
	let g_angle= degToRad(prop[n].g_angle) + Math.PI/2;  // 	 g_angle = big lines angle (radians);	
	g_angle-= Math.PI/4;    // => must decrease angle by this const = Math.PI/4   
	let big_cnt= prop[n].big_cnt;	
  //setBorder(n, g_angle);  debug function
	
	let c= document.getElementById('can'), x0= c.width/2, y0= c.height/2;		
	let rad = Math.floor( ( prop[n].big_size / 100 ) * Math.floor (0.5*Math.min(c.width, c.height)) );
	// 1. set raster of big (main) lines 
	let p1 = { x: x0 - rad * Math.cos(g_angle), y: y0 - rad * Math.sin(g_angle) };
	g_angle-= Math.PI/2;  // clockwise rotation 
	let p2 = { x: x0 - rad * Math.cos(g_angle), y: y0 - rad * Math.sin(g_angle) };
	
	// set start of each main stimulus line ("big" lines)
	let start= [];
	//start.push(p1);
	for (let j = 1; j <= big_cnt; j++ )   {	   // see wm6.js + wm7.js   					          
		let x= divLine(p1.x, p2.x, -j/(big_cnt-j+1));   // last arg = lambda = -m/n. example if step= 3: -1/3, -2/2, -3/1          
		let y= divLine(p1.y, p2.y, -j/(big_cnt-j+1));   // last arg = lambda = -m/n. example if step= 3: -1/3, -2/2, -3/1            	 
		let point = { 'x': x, 'y': y}; 
		start.push(point); 
	}		
	//start.push(p2);
	
	// set end of each main stimulus line ("big" lines)
	g_angle-= Math.PI/2;  // clockwise rotation 
	p1 = { x: x0 - rad * Math.cos(g_angle), y: y0 - rad * Math.sin(g_angle) };
	g_angle-= Math.PI/2;  // clockwise rotation 
	p2 = { x: x0 - rad * Math.cos(g_angle), y: y0 - rad * Math.sin(g_angle) };

	let end= [];	
	//end.unshift(p1);
	for (let j = 1; j <= big_cnt; j++ )   {	   // see wm6.js + wm7.js   					          
		let x= divLine(p1.x, p2.x, -j/(big_cnt-j+1));   // last arg = lambda = -m/n. example if step= 3: -1/3, -2/2, -3/1          
		let y= divLine(p1.y, p2.y, -j/(big_cnt-j+1));    // last arg = lambda = -m/n. example if step= 3: -1/3, -2/2, -3/1            	 
		let point = { 'x': x, 'y': y }; 
		end.unshift(point); 
	}	
	//end.unshift(p2);
	
	let c_angle = Math.atan2(end[0].y - start[0].y, end[0].x - start[0].x);  // angle in radians between two points of (first) big line. all big lines have same angle (at these moment)
	let r = getDistance(start[0].x, start[0].y, end[0].x, end[0].y);  // length (px) for 0. big line => distance for all fixed big lines is const
	
	let big = [];
	let up = start.length;			
	for (let i= 0; i < up; i++)  
	{			
		let angle = null, m_angle = null; // // real current angle in radians (only for odd lines)
		if ( i % 2 != 0 )   {  // shift-rotate each odd (1,3,5,..) big line
			//end[i].x= divLine(end[i].x, end[i-1].x, -1);  // half of distance between end of this and "previous" big line.
			//end[i].y= divLine(end[i].y, end[i-1].y, -1);
			end[i].x= end[i-1].x;  // x-end of each odd big line is now end of "previous" big line
			end[i].y= end[i-1].y;  // y-end of each odd big line is now end of "previous" big line
			angle = Math.atan2(end[i].y - start[i].y, end[i].x - start[i].x);  // real current angle in radians between two points of (first) big line. all big lines have same angle (at these moment)
			m_angle = Math.abs(angle-c_angle);  // maximum allowed angle (shift)
		}
		/* xs, ys = x,y coord of START point of each big line,
			 xe, ye = x,y coord of END point of each big line
			 c_angle = correct big (fixed) line atan2 angle,
			 angle = current big line atan2 angle	(or null)		
			 m_angle = maximum allowed atan2 angle (or null)
			 rad = fixed big line length (px) */
		let line = { 'xs' : start[i].x, 'ys' : start[i].y, 'xe' : end[i].x, 'ye' : end[i].y, 'c_angle' : c_angle, 'angle': angle, 'm_angle': m_angle, 'rad': r }; 
		big.push(line);
	}
	 
	stim[n].big = big;
	
	setSmall(n);

}


/* debug function. set borders of square region (4. points)
// n = item number. range 0..item_cnt-1
function setBorder(n, g_angle)  
{
	let c= document.getElementById('can'),  x0= c.width/2, y0= c.height/2;
	let rad = Math.floor( ( prop[n].big_size / 100 ) * Math.floor (0.5*Math.min(c.width, c.height)) );
	let border = []; 
	border[0]=  [ x0 + rad * Math.cos(g_angle), y0 + rad * Math.sin(g_angle) ]; 	 // 0 == x, 1 == y
	for (let i=1; i<=3; i++)  {
		g_angle-=Math.PI/2;
		border[i]=  [ x0 + rad * Math.cos(g_angle), y0 + rad * Math.sin(g_angle) ];	 // 0 == x, 1 == y		
	}			
	stim[n].border= border;  
}
*/


// set raster of small (oblique) lines
// n = item number. range 0..item_cnt-1
function setSmall(n)
{
	let big_cnt= prop[n].big_cnt;
	let small_cnt= prop[n].small_cnt;
	let i_angle= degToRad(prop[n].i_angle);
	
	let r = stim[n].big[0].rad; // full length of 0. big line (this line is fixed => not rotated)	
	let s_size = 0.5 * r / (big_cnt+1); // s_size => radius (length) for small lines => half of distance of big (main) lines / number of small lines
		
	let small= [];
	for (let i in stim[n].big)  
	{					 
		let p1 = { x: stim[n].big[i].xs, y : stim[n].big[i].ys };
		let p2 = { x: stim[n].big[i].xe, y : stim[n].big[i].ye };
	  let f_alpha = Math.atan2(p2.y - p1.y, p2.x - p1.x);  // angle in radians for this i (fixed or moved) line
		let ang = i % 2 == 0 ? f_alpha - i_angle : f_alpha + i_angle;   // let ang = i % 2 == 0 ? f_alpha - i_angle : f_alpha - i_angle + Math.PI/2;  
		for (let j = 1; j <= small_cnt; j++ )   {	   // see wm6.js + wm7.js   					          
			let x0 = divLine(stim[n].big[i].xs, stim[n].big[i].xe, -j/(small_cnt-j+1)) ;   // last arg = lambda = -m/n. example if step= 3: -1/3, -2/2, -3/1          			
			let y0 = divLine(stim[n].big[i].ys, stim[n].big[i].ye, -j/(small_cnt-j+1));    // x, y = start of small (oblique) line           	 					          	 					
			let point = { 'x0': x0, 'y0': y0, 'rad' : s_size, 'angle': ang}; // rad = polar radius(px), angle = angle of each small line (radian)
			small.push(point); 
		}				
	}
	stim[n].small= small;
}


// convert degrees (deg) to radians
function degToRad(deg)  {
	return (deg * Math.PI/180);   
}


//  convert radians (rad) to degrees
function radToDeg(rad)  {
	return (rad * 180 / Math.PI); 
} 
 
 
// isNaN("") = false: the empty string is converted to 0 which is not NaN                   
function isInt(n) {       
   return (!isNaN(n)) && (parseFloat(n) === parseInt(n, 10)) &&  (n % 1 === 0); 
} 


// trim input string: str
function trim(str) {
   return str.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
}


// unbiased shuffle algorithm: Fisher-Yates (aka Knuth) Shuffle.
function shuffle(arr) 
{
  let cur_index = arr.length, tmp, rnd_index;
  // While there remain elements to shuffle...
  while (0 !== cur_index) {    
    rnd_index = Math.floor(Math.random() * cur_index);     // Pick a remaining element...
    cur_index = cur_index - 1;   
    tmp = arr[cur_index];          // And swap it with the current element.
    arr[cur_index] = arr[rnd_index];
    arr[rnd_index] = tmp;
  }
  return arr;
}


// https://www.w3.org/2002/09/tests/keys.html	 
// https://stackoverflow.com/questions/17164239/add-delete-and-arrow-key-into-regular-expression 
// https://stackoverflow.com/questions/13196945/keycode-values-for-numeric-keypad/13196983
// all variables => basic textarea validator => onkeydown() => see html tag
function updateVar(event)
{
	event = event || window.event;
	let key = event.keyCode || event.which;
	if( key == 13 || key == 37 || key == 38 || key == 39 || key == 40 || key == 8 || key == 46 ) { // ENTER, Left / Up / Right / Down Arrow, Backspace, Delete keys
		return;  // Don't validate the input if ENTER, arrows, delete and backspace keys were pressed 
	}
	let is_digit = (key >= 48 && key <= 57) || (key >= 96 && key <= 105);
	if( !is_digit )  {
	event.returnValue = false;
	if(event.preventDefault) 
		event.preventDefault();
	}   
}


// return distance from point P1(x1,y1) to point P2(x2,y2)
function getDistance(x1, y1, x2, y2)  {
  return Math.sqrt( Math.pow(x1-x2, 2) + Math.pow(y1-y2, 2) ); // total line distance (px) 
}	


/* controls the keyboard during a task. It controls the size of the illusion, ie the dependent variables */
function onKeyDown(e)  
{
	if (is_task) 
	{
		e = e || window.e;   
		let key = e.keyCode;  	
		if ( key != 37 && key != 39 )  {
			return;  // arrow Left = 37. arrow Right = 39
		}		
		// change desired (rotation) moving steps here.. 
		let step = 1800;  // default step =  1/10°  or  0.1 degree
		if ( e.shiftKey ) {
			step = 180;  //  1°
		}
		else if ( e.ctrlKey )  {
		  step = 18000;   //  1/100° or 0.01 degree
		}
		let r = stim[item].big[1].angle;  // each odd (i=1,3,5,..) big lines have:  stim[item].big[i].angle = const
		if ( key == 37 ) { // arrow Left = 37
			r+= Math.PI/step; // increase angle by step (0.1, 1 or 0.01) degree
		}
		else { // arrow Right = 39
			r-= Math.PI/step; // decrease angle by step (0.1, 1 or 0.01) degree 
		}
		// first check new radius (coord)..
		if ( Math.abs( r - stim[item].big[1].c_angle ) > stim[item].big[1].m_angle ) {
			let elem = document.getElementById('nav_items');
			elem.textContent= 'Out of range. Press the opposite arrow key'; 
			elem.setAttribute('style', 'color:#f00;font-weight:bold');
			out_range= true;
			return;
		}
		// radius ok. update big (main, test) lines slope..
		let up= stim[item].big.length-1;			
		for (let i= 1; i <= up; i+=2)  
		{		// rotate each odd (1,3,5,..) big line	
			if ( key == 37 ) { 
				stim[item].big[i].angle+= Math.PI/step; // // increase angle by step degree
			}
			else { 
				stim[item].big[i].angle-= Math.PI/step; // decrease angle by step degree
			}
			stim[item].big[i].xe= stim[item].big[i].xs + stim[item].big[i].rad * Math.cos(stim[item].big[i].angle);  // update x-end of big line
			stim[item].big[i].ye= stim[item].big[i].ys + stim[item].big[i].rad * Math.sin(stim[item].big[i].angle);	 // update y-end of big line			  		
		}
		
		setSmall(item);  // update small lines
		mis[item]= false;
		drawStim(false);     // redraw scene (stimulus)
	}
} 


// delete warning on footer "Out of range. Press the opposite arrow key"
function onKeyUp(event)  
{
  if (is_task && out_range)  
	{
    let e = document.getElementById('nav_items');
		e.removeAttribute('style');
		if  (prg_type)  {    // update Progress bar => show current item count and remaining item count on Progress Bar                                                                                                                                                        
			e.textContent= 'Item= ' + (item + 1) + ' Remaining= ' + (item_cnt - item-1);  
		}
		else  {
			e.textContent= '';
		}
		out_range= false;
  }		
}


// draw (repaint) current stimulus
// footer => update footer (true) of false (only for calls from onKeyDown() event)            
function drawStim(footer) 
{    			
	let c= document.getElementById('can'), ctx= c.getContext('2d'), w= c.width, h= c.height;		
  ctx.fillStyle= back_clr;
	ctx.fillRect(0, 0, w, h);    // ctx.clearRect(0, 0, w, h); 
	
	ctx.strokeStyle= pen_clr; 

	// draw raster of main stimulus lines ("big" lines)	
	for (let i in stim[item].big) {
		ctx.beginPath();
		ctx.moveTo( stim[item].big[i].xs, stim[item].big[i].ys ); 
		ctx.lineTo( stim[item].big[i].xe, stim[item].big[i].ye );
		ctx.stroke();
	}
	
	// draw raster of small (oblique) lines - in each main (big) lines
	for (let i in stim[item].small)  {
		ctx.beginPath();
		ctx.moveTo( stim[item].small[i].x0 - stim[item].small[i].rad * Math.cos(stim[item].small[i].angle), stim[item].small[i].y0 - stim[item].small[i].rad * Math.sin(stim[item].small[i].angle) ); 
		ctx.lineTo( stim[item].small[i].x0 + stim[item].small[i].rad * Math.cos(stim[item].small[i].angle), stim[item].small[i].y0 + stim[item].small[i].rad * Math.sin(stim[item].small[i].angle) ); 
		ctx.stroke();
	}
	
	//getAngleBetweenSmall();  // debug
	
	if (is_circle)  {  // draw circle around all stimuli
		let rad = Math.floor( ( prop[item].big_size / 100 ) * Math.floor (0.5*Math.min(w, h)) );
		rad -= Math.round(pen_width/2);
		ctx.beginPath();
		ctx.arc(w/2, h/2, rad, 0, 2*Math.PI);
		ctx.stroke();	
	}
	
	//drawBorder(ctx);  // debug

	if ( demo && !mis[item] )  
	{	
		ctx.strokeStyle= pen_clr_c;
		let up= stim[item].big.length;
		for (let i= 1; i< up; i+= 2  ) {
			ctx.beginPath();
			ctx.moveTo( stim[item].big[i].xs, stim[item].big[i].ys ); 
			ctx.lineTo( stim[item].big[i].xs + stim[item].big[i].rad * Math.cos(stim[item].big[i].c_angle), stim[item].big[i].ys + stim[item].big[i].rad * Math.sin(stim[item].big[i].c_angle) );
			ctx.stroke();
		}
		calcError(item);
		document.getElementById('nav_items').textContent= 'Absolute error= ' + dv[item][0].toFixed(3) + '. Relative error= ' + dv[item][1].toFixed(3);		// show error on middle of footer
	}		
	
	if (footer) {
		updateFooter();
	}
}


/* debug
// draw rect border around all stimuli 
function drawBorder(ctx)
{
	//ctx.strokeStyle= '#f00'; 
	ctx.beginPath();
	ctx.moveTo( stim[item].border[0][0], stim[item].border[0][1] );
	for (let j= 1; j<= 3; j++)  {
		ctx.lineTo( stim[item].border[j][0], stim[item].border[j][1] );
	}
	ctx.closePath();
	ctx.stroke();  
}
*/


/* debug
// return angle between even and odd lines in current item
function getAngleBetweenSmall()
{
	let x0= stim[item].small[0].x0, y0 = stim[item].small[0].y0, rad = stim[item].small[0].rad, angle= stim[item].small[0].angle; // 1. fixed small line
	let x1=  x0 - rad * Math.cos(angle), y1= y0 - rad * Math.sin(angle); 
	let x2 = x0 + rad * Math.cos(angle), y2= y0 + rad * Math.sin(angle);
	let m1 = ( y2 - y1 ) / ( x2 - x1 );
	let alpha1 = Math.atan2(y2 - y1, x2 - x1);  // angle in radians for 0. fixed line
  
	let n= prop[item].small_cnt;  // first small line in second big line
	x0= stim[item].small[n].x0;
	y0 = stim[item].small[n].y0; 
	rad = stim[item].small[n].rad;
	angle= stim[item].small[n].angle;  // data from 1. fixed moving line
	x1=  x0 - rad * Math.cos(angle), y1= y0 - rad * Math.sin(angle); 
	x2 = x0 + rad * Math.cos(angle), y2= y0 + rad * Math.sin(angle);
	let m2 = ( y2 - y1 ) / ( x2 - x1 );
	let alpha2 = Math.atan2(y2 - y1, x2 - x1);  // angle in radians for 0. fixed line	
	let alpha = Math.atan( Math.abs( ( m2 - m1) / ( 1 + m1 * m2 ) ) );         //tan alpha = | ( m2 - m1) / ( 1 + m1 * m2 ) |	
}
*/


/* calculate dependent variable: absolute and relative error.
  i => item number	
  data for each stim[i].big[j]=
  {	xs, ys = x,y coord of START point of each big line,
		xe, ye = x,y coord of END point of each big line
		c_angle = correct big (fixed) line atan2 angle,
		angle = current big line atan2 angle	(or null)		
		m_angle = maximum allowed atan2 angle (or null)
		rad = fixed big line length (px) } 		*/
function calcError(i)  
{
	let c_angle= stim[i].big[0].c_angle;  // correct angle (all are equal)
	let angle= stim[i].big[1].angle;  // current angle for moved line (all odd lines have same angle)
  dv[i][0]= radToDeg(Math.abs(c_angle - angle));  // absolute errror in degrees. Note: always is >= 0 !! 
	dv[i][1]=  100 * ( Math.abs(c_angle - angle) / stim[i].big[1].m_angle ) ;  // relative error in "percents". approx range = [0..100] 
}


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
  if (document.getElementById('variable').selectedIndex > -1 )  {
		variable = document.getElementById('variable').value;
  }
	
	if (document.getElementById('iter').validity.valid)  {  
	  iter = parseInt(document.getElementById('iter').value, 10);			
  } 	
	
	if (validColor(document.getElementById('back_clr').value))  {		
		back_clr= document.getElementById('back_clr').value;
  }
	
	if (validColor(document.getElementById('pen_clr').value))  {		
		pen_clr= document.getElementById('pen_clr').value;
  }
	
	if (validColor(document.getElementById('pen_clr_c').value))  {		
		pen_clr_c= document.getElementById('pen_clr_c').value;
  }
	
	if (document.getElementById('pen_width').validity.valid)  {  
	  pen_width = Math.round(parseInt(document.getElementById('pen_width').value, 10));			
  }

	is_circle= document.getElementById('is_circle').checked ? true : false;
  
	is_back= document.getElementById('is_back').checked ? true : false;      
  	
	prg_type= document.getElementById('prg_type').checked ? true : false;   
		
  demo= document.getElementById('demo').checked ? true : false;	
	
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;
  
}


// test color for validity (from string)
// https://stackoverflow.com/questions/8027423/how-to-check-if-a-string-is-a-valid-hex-color-representation/8027444
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// goto previous item (stimulus)
function priorItem()  
{      
  if (item > 0)  {
    item--;  
    drawStim(true);     // draw again previous stimulus
  }         
}


// goto next item (stimulus)
function nextItem()  
{    		
	item++;					
	if (item < item_cnt)  {         		 
    drawStim(true);     // draw next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd(false);  // end of task if item >= item_cnt 
  }
}


/* divide line 
  first => first point (P1) 
  second => second point (P2)
  n => (distance(P1-Q)/ distance(Q-P2)) = lambda = -m/n = negative ratio length: P1-Q/Q-P2
  use n= -1 for half divide line (50%) 
	returns: coordinate of point (Q) between first and second point   */   
function divLine(first,second,n)  {     
  return (first-n*second)/(1-n);   // Math.round((first-n*second)/(1-n));
}


// update footer when show each stimulus in task. update buttons and update progress bar
function updateFooter()
{  
	document.getElementById('prior').style.display= item > 0 && is_back ? 'inline': 'none';  // set footer navigation
	let btn= document.getElementById('next');     
	if (item == item_cnt-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update 'Next item' image on 'right arrow' button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Go forward';
	}						
  if (prg_type)  {    // update Progress bar => show current item count and remaining item count on Progress Bar                                                                                                                                                        
		let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent= 'Item= ' + (item + 1) + ' Remaining= ' + (item_cnt - item-1);  
	}
}


// return distance in px between small lines in i-item
// i => item number
function getSmallDistance(i)
{
 // all distance between fixed (even line 0,2,4,..) in i-item are const
 //stim[i].small[j] => { 'x0': x0, 'y0': y0, 'rad' : s_size, 'angle': ang}; // T(x0,y0) is intersection of big and small line
 return getDistance( stim[i].small[0].x0, stim[i].small[0].y0, stim[i].small[1].x0, stim[i].small[1].y0 ); 
}


// return fixed line length in px for i-item
// i => item number
function getBigDistance(i)
{
 return stim[i].big[0].rad;   // all distance between fixed (even line 0,2,4,..) in i-item are const
 // or return d= getDistance( stim[i].big[0].xs, stim[i].big[0].ys, stim[i].big[0].xe, stim[i].big[0].ye );  // 1. fixed line length. all distance between fixed (even line 0,2,4,..) in i-item are const 
}

 
// end of task (research)
// if (is_error== true) => page (window) was resized during task. Show only html error warning
function theEnd(is_error)  
{     
	is_task= false;
	
	document.getElementById('task_section').style.display= 'none';	
  document.getElementById('end').style.display= 'block';   
  document.getElementById('can').style.display= 'none';  
 
 
	if (is_error) {  // Show only html error warning
		document.getElementById('error').style.display= 'block';
		document.getElementById('result_wrapper').style.display= 'none';
		document.getElementById('score').textContent= '';
		return;
	}
	
  document.getElementById('error').style.display= 'none';
	document.getElementById('result_wrapper').style.display= 'block';
	
	const D = 3;  // number of decimals
  let rows = [], header = [], var_name, blank= 0, tmp;  // blank = missing (not-moved) items count, tmp= variable value selected by user 
/*	
  abs_v = array of non-missing absolute errors (units: degrees °)
	rel_v = array of non-missing relative errors (units: percents from max allowed error)
	var_v = array of independent (abscise) variable. if necessary, converted to px (for variable= '3' or 'variable= '5'). only for non-missing items.
	var_g => dependent variable group by var_v value (only non-missing items)   			
*/	
	let abs_v = [], rel_v = [], var_v = [], var_g= {}; 
		// variable => Variable selector: '1' = Intersect angle; '2' = Big angle '3' = Small lines; '4' = Big lines; '5' = Big length   
	let e= document.getElementById('variable');  
	var_name = e.options[e.selectedIndex].textContent;	
	if (variable == '3' || variable == '5')  {  // 6. columns
		header= [ 'Item', var_name, var_name + '-px', 'Absolute error dg', 'Relative error %', 'Moved' ];
	}   // degree ASCII code alt + 0176:  ° ° °
	else   { 	// 5. columns
	  header= [ 'Item', var_name,                   'Absolute error dg', 'Relative error %', 'Moved' ];
	}

	for (let i = 0; i < item_cnt; i++)  
	{	
		// variable => '1' = Intersect angle; '2' = Angle of big lines; '3' = Number of small lines; '4' = Number of big lines; '5' = Size of big lines   		
		// prop[i] = { 'i_angle': i_angle[0], 'g_angle': g_angle[0], 'small_cnt': small_cnt[i], 'big_cnt': big_cnt[0], 'big_size': big_size[0] }; 
		switch (variable)  {				
			case '1':				
				var_v.push (prop[i].i_angle);
				tmp= prop[i].i_angle;   // tmp => variable value selected by user (in this case equal var_v)
				break;
			case '2':
				var_v.push ( prop[i].g_angle );
				tmp= prop[i].g_angle;   
				break;
			case '3':				
				var_v.push ( getSmallDistance(i) );  // distance between small lines (units: pixels)
				tmp = prop[i].small_cnt;			//	tmp => variable value selected by user => 2. column in table1 
				break;
			case '4':				
				var_v.push (prop[i].big_cnt);  
				tmp = prop[i].big_cnt;				
				break;
			case '5':
				var_v.push ( getBigDistance(i) ); // distance between big lines (units: pixels)
				tmp= prop[i].big_size;         // tmp => 2. column in table1 (variable value selected by user)
				break;					
		}		
		let last = var_v[var_v.length-1];
		// for decriptive statistics set group by variable value (table2 + table3)..
		let key = last.toFixed(D);   // key => last one pushed into var_v
		if ( typeof var_g[key]  === 'undefined' )  {
			var_g[key] = {   
				'absolute'  : [],
				'relative' : []
			}; 
		}

    if ( mis[i] == true )  
		{
			blank++;	
		  if (variable == '3' || variable == '5')  { 		//  6. columns 				   			
				rows[i]= [ parseInt(i+1,10), tmp,  last.toFixed(D),  '', '', 'No' ];  // add 3. column variable converted to px 
			}
			else  {  //  5. columns	   			
				rows[i]= [ parseInt(i+1,10), tmp,  '', '', 'No' ];  	// tmp => variable value selected by user
			}
		}
		else  
		{
			calcError(i); 
			
		  abs_v.push ( dv[i][0] );  // absolute error (°)
		  rel_v.push ( dv[i][1] );  // relative error (%)
      if (variable == '3' || variable == '5')  { 		//  6. columns 
				rows[i]= [ parseInt(i+1,10), tmp,  last.toFixed(D),   dv[i][0].toFixed(D), dv[i][1].toFixed(D), 'Yes' ]; 
			}
			else  {  // 5. columns  				   			
				rows[i]= [ parseInt(i+1,10), tmp,                     dv[i][0].toFixed(D), dv[i][1].toFixed(D), 'Yes' ]; 
			}
			
			var_g[key].absolute.push( dv[i][0] );  // error ° => abs_v
			var_g[key].relative.push( dv[i][1] );  // error % => rel_v
		}
	}  // for (let i = 0; i < item_cnt; i++)
		
  createTableArray(document.getElementById('table1'), rows, header); 
 
  if  (variable == '3' ) {
    document.getElementById('var_name_px').textContent= var_name + '-px = For all statistics, the number of small lines is converted to the distance between small lines (within each big line) in pixels.';
		document.getElementById('var_name_px').style.display= 'block';	
	}		
  else if  (variable == '5' ) {
    document.getElementById('var_name_px').textContent= var_name + '-px = For all statistics, user-selected big line sizes are converted to pixels.';
	  document.getElementById('var_name_px').style.display= 'block';	
	}		
	else  {
    document.getElementById('var_name_px').style.display= 'none';	
	}
 
	// table2 => Descriptive statistics for Absolute error (degrees °) grouped by selected variable and total
  document.getElementById('table2_title').textContent= (variable == '3' || variable == '5') ? 'Absolute error (degrees). Descriptive statistics grouped by variable ' + var_name + '-px and total:' : 'Absolute error (degrees). Descriptive statistics grouped by variable ' + var_name + ' and total:' 
	
	let arr, n, m, sd;
	rows = [];  // reset table rows (for table2)
	for (let key in var_g)  {
		arr= var_g[key].absolute;  // grouped by variable for this key
		n= arr.length;
		m =  ( n > 0 ) ? Stat.m(arr).toFixed(D) : '';     //  Absolute errors. mean only for non-missing items in task
		sd = ( n > 1 ) ? Stat.sd(arr, m).toFixed(D): '';  //  Absolute errors. standard deviation only for non-missing items in task  	 	
		rows.push( [key, m, sd, n] );  
	}
	// totals (degrees)
	n= abs_v.length;
	m =  ( n > 0 ) ? Stat.m(abs_v).toFixed(D) : '';    // Total absolute Errors (°). mean only for non-missing items in task
	sd = ( n > 1 ) ? Stat.sd(abs_v, m).toFixed(D): ''; // Total absolute Errors (°). standard deviation only for non-missing items in task  	 	
  rows.push( ['Total', m, sd, n] );  // last row
	let data = [ m, sd ];  // index 0 => mean for absolute error, index 1 => stand. deviation for absolute error
	header= [var_name, 'Average', 'Standard Deviation', 'N']; // header		
	createTableArray(document.getElementById('table2'), rows, header); 
		
	// table3 => Descriptive statistics for Relative error (%) grouped by selected variable and total
	document.getElementById('table3_title').textContent= (variable == '3' || variable == '5') ? 'Relative error (%). Descriptive statistics grouped by variable ' + var_name + '-px and total:' : 'Relative error (%). Descriptive statistics grouped by variable ' + var_name  + ' and total:';
	rows = [];  // reset table rows (for table3)
	for (let key in var_g)  {
		arr= var_g[key].relative;  // grouped by variable for this key
		n= arr.length;
		m =  ( n > 0 ) ? Stat.m(arr).toFixed(D) : '';     //  Relative errors (%). mean only for non-missing items in task
		sd = ( n > 1 ) ? Stat.sd(arr, m).toFixed(D): '';  //  Relative errors (%). standard deviation only for non-missing items in task  	 	
		rows.push( [key, m, sd, n] );  
	}
  // totals	(%)
  n= rel_v.length;
	m =  ( n > 0 ) ? Stat.m(rel_v).toFixed(D) : '';     // Total Relative errors (%). mean only for non-missing items in task
	sd = ( n > 1 ) ? Stat.sd(rel_v, m).toFixed(D): '';  // Total Relative errors (%). standard deviation only for non-missing items in task  	 	
  rows.push( ['Total', m, sd, n] );  // last row	
	data.push(m);   // index 2 => mean for Relative error (%)
	data.push(sd);  // index 3 => stand. deviation for Relative error (%)
	createTableArray(document.getElementById('table3'), rows, header); 

  if (variable == '3' || variable == '5')  { 
		var_name+= '-px';  // for table4 + setStorage()
	}
	// table4 => calculate inferential statistics (pearson r, sperman ro, t-arguments, t-distribution probabilities)	
	n= abs_v.length;  // or  n= rel_v.length; 
	header=  ['Variables', 'Spearman-ro', 't-value ro', 'df', 'p-ro (two-sided)',  'Pearson r', 't-value r', 'df', 'p-r (two-sided)' ]; // header 
	rows=[];    // reset table rows (for table4)
	for (let j= 0; j<=1; j++)
	{
		
		let y=[], vars;
		if ( j == 0 )  {
			y= abs_v;   // errors in °
			vars= var_name + ' - Absolute error';
		}
		else  {
			y= rel_v;    // error in percents
			vars= var_name + ' - Relative error';
		}	
		 
		let row= [ vars, '', '', '', '', '', '', '', '' ];  // range 0..8
		if  ( n >= 3 ) 
		{			
			let ro= Stat.r_spearman(var_v, y);     // spearman's rank correlation
			if (!isNaN(ro) && isFinite(ro) )  // if no variance in both arrays (sd= 0) => Infinity or Nan
			{ 
				row[1]= ro.toFixed(D);  
				let t_ro= ro * Math.sqrt ( ( n-2 ) / (1 - ro * ro) );  // t-argument for spearman correlation 		
        if (!isNaN(t_ro) && isFinite(t_ro))  
				{
					let err=[0];  // reset error flag indicator (must be 0)
					let p_ro= Stat.prncst ( Math.abs(t_ro), n-2, 0, err[0]); // t-argument, df, 0, error
					p_ro=  2 * (1 - p_ro);  // Propability p (two-sided) => tested ok for agument n= 3..1000 with: https://www.psychometrica.de/correlation.html 
					row[2]= t_ro.toFixed(D);
					row[3]= n-2;
					if (err[0] == 0) { 
						row[4]= p_ro.toFixed(D); // propability Spearman ro (two-sided)
					}
				}
			}
			
			let r= Stat.r(var_v, y);   // Pearson r correlation
			if (!isNaN(r) && isFinite(r))    // if no variance in both arrays (sd= 0) => Infinity or Nan
			{
				row[5]= r.toFixed(D);  
				let t_r=  r * Math.sqrt ( ( n-2 ) / (1 - r * r) );  // t-argument for pearson correlation
				if (!isNaN(t_r) && isFinite(t_r)) 
				{					
					let err=[0];  // reset error flag indicator (must be 0)
					let p_r= Stat.prncst ( Math.abs(t_r), n-2, 0, err[0]);  // t-argument, df, 0, error
					p_r= 2 * (1 - p_r);  // Propability p (two-sided) => tested ok for agument n= 3..1000 with: https://www.psychometrica.de/correlation.html 
					row[6]= t_r.toFixed(D);
					row[7]= n-2;
					if (err[0] == 0) { 
						row[8]= p_r.toFixed(D);  // p-r (two-sided)
					}					 
				}
			}			
		}  //  if ( n >= 3 ) 
		
		rows.push( row );  // 9 x columns
		
		data.push(row[5]);  // pearson r  (index 4 or 7)
		data.push(row[6]);  // t-value r  (index 5 or 8)
		data.push(row[8]);  // p-r (two-sided) => index (6 or 9)
	
	}		// for (let j= 0; j<=1; j++)
	
	if ( n >= 3 ) {
	  data.push(n - 2);    // df => is equal for degrees and %. => or: abs_v.length - 2 or: n= rel_v.length - 2;
	}
  else  {
		data.push('');  // df => minimum is 3 non-missing items
	}
		
	createTableArray(document.getElementById('table4'), rows, header); 
  
	n= abs_v.length;  // non-missing items (absolute errors)
	document.getElementById('inf_warning').style.display= n >= 3 ? 'block' : 'none';	
	
  // update 'score' label => above "Repeat" button
	switch (n)  {
		case 0: 
			document.getElementById('score').textContent= 'Descriptive statistics cannot be calculated. Missing items= ' + blank;
			break;
		case 1: 
			document.getElementById('score').textContent= 'Total average Absolute error (degree)= ' + data[0] + '. Standard deviation cannot be calculated. Missing items= ' + blank;
			break;	
		default:
		document.getElementById('score').textContent= 'Total average Absolute error (degree)=  ' + data[0] + '. Standard Deviation= ' + data[1] + '. Missing items= ' + blank; 
	}
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
	
	let var_values= '';
  for (let key in var_g)  {
    var_values += key + ';';
  }
	
	try 
  {
    if (store) {
			setStorage(var_name, var_values, blank, data);  
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  } 
	
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if (dd<10) { dd='0'+dd; } 
	if (mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:
	 var_name => variable name (string description)
	 var_values => list of variable values (if necessary they are recoded in px). separator in list is ";"  
	 blank => number of missing items in task 
	 data => descriptive and inferential statistics (index range: 0..10). for details see theEnd()  */ 
function setStorage(var_name, var_values, blank, data)  
{    
	
  // variable => '1' = Intersect angle; '2' = Big angle '3' = Small lines; '4' = Big lines; '5' = Big length 
	// prop[i] = { 'i_angle': i_angle[0], 'g_angle': g_angle[0], 'small_cnt': small_cnt[i], 'big_cnt': big_cnt[0], 'big_size': big_size[0] }; 
	
	let oldItems = JSON.parse(localStorage.getItem('zi-v1.0')) || []; 
  let id= oldItems.length + 1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),     //new Date(),  		   
			'variable' : var_name,
			'iteration' : iter,
			'intersect-angle' : variable == '1' ? var_values :  prop[0].i_angle,
			'big-angle' : variable == '2' ? var_values :  prop[0].g_angle,
			'small-lines' : variable == '3' ? var_values :  prop[0].small_cnt,
			'big-lines' : variable == '4' ? var_values :  prop[0].big_cnt,	
      'big-length' : variable == '5' ? var_values : prop[0].big_size,	
			'pen-w': pen_width,
			'is-circle' : is_circle ? 1 : 0, 	
			'is-back' : is_back ? 1 : 0,
			'demo' : demo ? 1 : 0, 			
			'items' : item_cnt,
			'missing'  : blank, 			
			'm-abs'  : data[0], 	
			'sd-abs' : data[1],
      'm-rel'  : data[2], 	
			'sd-rel' : data[3],
			'r-abs'  : data[4],
			't-abs'  : data[5],
			'p-abs'  : data[6],			
			'r-rel'  : data[7],
			't-rel'  : data[8],
			'p-rel'  : data[9],			
			'df'     : data[10]
  };

  oldItems.push(newItem);

  localStorage.setItem('zi-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems); 
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
          col.push(key);
        }
    }
  } 
  let table = document.createElement('table');
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement('th');      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= '';
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement('table');  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement('th');      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('zi-v1.0')  ?  JSON.parse(localStorage.getItem('zi-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    //alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('zi-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('zi-v1.0');
	if (!data)  {
    //alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// copied from https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('zi-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved cumulative data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object (div or table)
function newWinLast(id)
{ 	
	let ind = document.getElementById('last_data_sel').selectedIndex;
	ind++;
	let e = document.getElementById('table' + ind);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex, tbl;
  switch (ind) {
    case 0:
			tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
			break;		
    case 1:
			tbl = document.getElementById('table2').getElementsByTagName('table')[0];  
			break;
		case 2:
			tbl = document.getElementById('table3').getElementsByTagName('table')[0];  
			break;			
		default:
			tbl = document.getElementById('table4').getElementsByTagName('table')[0];  // note: id='table4' is html div element
			break;	
	}		
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}


// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Classes
// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Classes/static
// http://www.rosettacode.org/mw/index.php?title=Special%3ASearch&search=integration&go=Go&fulltext=1
class Stat 
{ 
		
	// return arithmetic mean (mean, average) of input array (arr)
  // no validation => input array must be in valid number format	
	static m(arr)  
	{	
		let len= arr.length, sum= 0;
		for (let i=0; i<len; i++) {
			sum+= arr[i];
		}
		return (sum/arr.length);
	}	


	// return Sample Standard Deviation (with Bessel's correction) => N-1
	// no validation => input array must be in valid number format
	// arr = array argument;  mean = average for input arr 
	static sd(arr, mean) 
	{
		let len= arr.length, sum= 0, res;
		for (let i=0; i<len; i++) {
			sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
		}
		res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
		return res;
	}
	
	
	// https://www.psychometrica.de/correlation.html
	// return Pearson correlation coefficient [ Pearson's r, Pearson product-moment correlation coefficient, bivariate correlation ]
	//  x, y = arrays 
	static r(x,y)
	{
		if (x.length != y.length) {
			return;  // undefined
		}
		let cov= 0, n, m1, m2, sd1, sd2;
		n= x.length-1;     
		m1= this.m(x);          //MeanAndStdDev(a, M1, SD1);
		sd1= this.sd(x,m1);
		m2= this.m(y);
		sd2= this.sd(y,m2);    //MeanAndStdDev(b, M2, SD2);		
		for (let i=0; i <= n; i++) {
			cov= ( x[i] - m1 ) * ( y[i] - m2 ) + cov;  // covariance
    }
		return ( cov / ( n * sd1 * sd2 ) );   // denominator = n-1 (Bessel's correction)		
	}
	
	
	// return covariance
  // x, y = arrays   
	static cov(x,y)
	{
		if (x.length != y.length) {
			return;  // undefined
		}
		let cov= 0, n, m1, m2, sd1, sd2;
		n= x.length-1;     
		m1= this.m(x);          
		sd1= this.sd(x,m1);
		m2= this.m(y);
		sd2= this.sd(y,m2);    		
		for (let i=0; i <= n; i++) {
			cov= ( x[i] - m1 ) * ( y[i] - m2 ) + cov  ;  // covarijanca
    }
		return cov;   // denominator = n-1 (Bessel's correction)		
	}


	// https://www.statisticshowto.com/probability-and-statistics/correlation-coefficient-formula/spearman-rank-correlation-definition-calculate/   
	// https://mathcracker.com/spearman-correlation-calculator
	// compute Spearman's rank correlation coefficient
	// x, y = arrays
	static r_spearman(x, y)
	{ 
		if (x.length != y.length) {
			return;  // undefined
		}
		const N = x.length;  // n is the number of observations
		let xr = x.slice();  // must be
		xr= this.rankify(xr, xr.length);   // get (tied) ranks for 1. array values
		
		let yr = y.slice();  // must be
		yr= this.rankify(yr, yr.length);	  // get (tied) ranks for 2. array values
		
		let mx= this.m(xr);  // mean rank for 1. array
		let my= this.m(yr);  // mean rank for 2. array
		
		let sxy= 0, sx= 0, sy= 0;
		for (let i= 0; i < N; i++)  
		{
			sxy+= ( xr[i] - mx ) * ( yr[i] - my); 
			sx+= Math.pow( ( xr[i] - mx ), 2);
			sy+= Math.pow( ( yr[i] - my ), 2);
		}
		sxy= sxy/N;
		sx= sx/N;
    sy= sy/N;
    let res = sxy / Math.sqrt(sx * sy);
		return res;	
	}
	
	
	// https://www.geeksforgeeks.org/rank-elements-array/
	// function to find tied ranks
	// A => input array;  n => input array length
	static rankify(A, n) 
	{	
		let R = [...Array(n)];  // Rank Vector
		// Sweep through all elements in A for each
		// element count the number of less than and
		// equal elements separately in r and s.
		for (let i = 0; i < n; i++) 
		{
			let r = 1, s = 1;
			for (let j = 0; j < n; j++) {
				if (j != i && A[j] < A[i]) {
					r += 1;
				}
				if (j != i && A[j] == A[i]) {
					s += 1;
				}
			}
			// Use formula to obtain rank
			R[i] = parseFloat(r + parseFloat(s - 1) / parseFloat(2));
		}
		return R;
	}
 
   
	// https://people.sc.fsu.edu/~jburkardt/cpp_src/asa005/asa005.cpp
	// https://people.sc.fsu.edu/~jburkardt/cpp_src/asa005/asa005.html
	// https://people.math.sc.edu/Burkardt/cpp_src/toms291/toms291.cpp
	// https://people.math.sc.edu/Burkardt/c_src/toms291/toms291.html


	//****************************************************************************80

	static alnorm ( x, upper )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    ALNORM computes the cumulative density of the standard normal distribution.
	//
	//  Licensing:
	//
	//    This code is distributed under the GNU LGPL license. 
	//
	//  Modified:
	//
	//    17 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by David Hill.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    David Hill,
	//    Algorithm AS 66:
	//    The Normal Integral,
	//    Applied Statistics,
	//    Volume 22, Number 3, 1973, pages 424-427.
	//
	//  Parameters:
	//
	//    Input, double X, is one endpoint of the semi-infinite interval
	//    over which the integration takes place.
	//
	//    Input, bool UPPER, determines whether the upper or lower
	//    interval is to be integrated:
	//    .TRUE.  => integrate from X to + Infinity;
	//    .FALSE. => integrate from - Infinity to X.
	//
	//    Output, double ALNORM, the integral of the standard normal
	//    distribution over the desired interval.
	//
	{
		let a1 = 5.75885480458;
		let a2 = 2.62433121679;
		let a3 = 5.92885724438;
		let b1 = -29.8213557807;
		let b2 = 48.6959930692;
		let c1 = -0.000000038052;
		let c2 = 0.000398064794;
		let c3 = -0.151679116635;
		let c4 = 4.8385912808;
		let c5 = 0.742380924027;
		let c6 = 3.99019417011;
		let con = 1.28;
		let d1 = 1.00000615302;
		let d2 = 1.98615381364;
		let d3 = 5.29330324926;
		let d4 = -15.1508972451;
		let d5 = 30.789933034;
		let ltone = 7.0;
		let p = 0.398942280444;
		let q = 0.39990348504;
		let r = 0.398942280385;
		let utzero = 18.66;
		let value;
		let y;
		let z = x;

		let up = upper;   // bool
		z = x;

		if ( z < 0.0 )
		{
			up = !up;
			z = - z;
		}

		if ( ltone < z && ( ( !up ) || utzero < z ) )
		{
			if ( up ) {
				value = 0;
			}
			else  {
				value = 1;
			}
			return value;
		}

		y = 0.5 * z * z;

		if ( z <= con )
		{
			value = 0.5 - z * ( p - q * y 
				/ ( y + a1 + b1 
				/ ( y + a2 + b2 
				/ ( y + a3 ))));
		}
		else
		{ 
			value = r * Math.exp( - y )   
				/ ( z + c1 + d1 
				/ ( z + c2 + d2 
				/ ( z + c3 + d3 
				/ ( z + c4 + d4 
				/ ( z + c5 + d5 
				/ ( z + c6 ))))));
		}

		if ( !up )  {
			value = 1 - value;
		}

		return value;
	}


	//****************************************************************************80

	static prncst ( st, idf, d, ifault )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    PRNCST computes the lower tail of noncentral T distribution.
	//
	//  Licensing:
	//
	//    This code is distributed under the GNU LGPL license. 
	//
	//  Modified:
	//
	//    26 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by BE Cooper.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    BE Cooper,
	//    Algorithm AS 5:
	//    The Integral of the Non-Central T-Distribution,
	//    Applied Statistics,
	//    Volume 17, Number 2, 1968, page 193.
	//
	//  Parameters:
	//
	//    Input, double ST, the argument.
	//
	//    Input, int IDF, the number of degrees of freedom.
	//
	//    Input, double D, the noncentrality parameter.
	//
	//    Output, int *IFAULT, error flag.
	//    0, no error occurred.
	//    nonzero, an error occurred.
	//
	//    Output, double PRNCST, the value of the lower tail of
	//    the noncentral T distribution.
	//
	//  Local Parameters:
	//
	//    Local, double G1, 1.0 / sqrt(2.0 * pi)
	//
	//    Local, double G2, 1.0 / (2.0 * pi)
	//
	//    Local, double G3, sqrt(2.0 * pi)
	//
	{
		let a, ak, b,  da, drb,  fk, fkm1, fmkm1, fmkm2, ioe, k, rb, sum, value;   
		let emin = 12.5;
		let g1 = 0.3989422804;
		let g2 = 0.1591549431;
		let g3 = 2.5066282746;
		let f = idf;   //( double ) ( idf );
	//
	//  For very large IDF, use the normal approximation.
	//
		if ( 100 < idf )
		{
			ifault = 1;

			a = Math.sqrt ( 0.5 * f ) * Math.exp ( this.lgamma ( 0.5 * ( f - 1 ) ) 
				- this.lgamma ( 0.5 * f ) ) * d;

			value = this.alnorm ( ( st - a ) / Math.sqrt ( f * ( 1 + d * d ) 
				/ ( f - 2 ) - a * a ), false );
			return value;
		}

		ifault = 0;
		ioe = ( idf % 2 );
		a = st / Math.sqrt ( f );
		b = f / ( f + st * st );
		rb = Math.sqrt ( b );
		da = d * a;
		drb = d * rb;

		if ( idf == 1 )
		{
			value = this.alnorm ( drb, true ) + 2 * this.tfn ( drb, a );
			return value;
		}

		sum = 0;

		if ( Math.abs ( drb ) < emin )
		{
			fmkm2 = a * rb * Math.exp ( - 0.5 * drb * drb ) 
			* this.alnorm ( a * drb, false ) * g1;
		}
		else
		{
			fmkm2 = 0;
		}

		fmkm1 = b * da * fmkm2;

		if ( Math.abs ( d ) < emin )
		{
			fmkm1 = fmkm1 + b * a * g2 * Math.exp ( - 0.5 * d * d );
		}

		if ( ioe == 0 )
		{
			sum = fmkm2;
		}
		else
		{
			sum = fmkm1;
		}

		ak = 1;
		fk = 2;

		for ( k = 2; k <= idf - 2; k = k + 2 )
		{
			fkm1 = fk - 1;
			fmkm2 = b * ( da * ak * fmkm1 + fmkm2 ) * fkm1 / fk;
			ak = 1 / ( ak * fkm1 );
			fmkm1 = b * ( da * ak * fmkm2 + fmkm1 ) * fk / ( fk + 1 );

			if ( ioe == 0 )
			{
				sum += fmkm2;
			}
			else
			{
				sum += fmkm1;
			}
			ak = 1 / ( ak * fk );
			fk += 2;
		}

		if ( ioe == 0 )
		{
			value = this.alnorm ( d, true ) + sum * g3;
		}
		else
		{
			value = this.alnorm ( drb, true ) + 2 * ( sum + this.tfn ( drb, a ) );
		}

		return value;
	}
	//****************************************************************************80


	// https://people.math.sc.edu/Burkardt/c_src/toms291/toms291.html
	// https://people.math.sc.edu/Burkardt/cpp_src/toms291/toms291.cpp
	//****************************************************************************80

	static lgamma ( x )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    ALOGAM computes the logarithm of the Gamma function.
	//
	//  Modified:
	//
	//    22 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by Malcolm Pike, David Hill.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    Malcolm Pike, David Hill,
	//    Algorithm 291:
	//    Logarithm of Gamma Function,
	//    Communications of the ACM,
	//    Volume 9, Number 9, September 1966, page 684.
	//
	//  Parameters:
	//
	//    Input, double X, the argument of the Gamma function.
	//    X should be greater than 0.
	//
	//    Output, double lgamma, the logarithm of the Gamma
	//    function of X.
	//
	{
		let f, value, y, z;

		if ( x <= 0 || isNaN(x) )  {
			return NaN;
		}

		y = x;

		if ( x < 7 )
		{
			f = 1;
			z = y;

			while ( z < 7 )
			{
				f = f * z;
				z++; 
			}
			y = z;
			f = - Math.log ( f );
		}
		else
		{
			f = 0;
		}

		z = 1 / y / y;

		value = f + ( y - 0.5 ) * Math.log ( y ) - y 
			+ 0.918938533204673 + 
			((( 
			- 0.000595238095238   * z 
			+ 0.000793650793651 ) * z 
			- 0.002777777777778 ) * z 
			+ 0.083333333333333 ) / y;

		return value;
	}

	//****************************************************************************80

	static tfn ( x, fx )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    TFN calculates the T-function of Owen.
	//
	//  Licensing:
	//
	//    This code is distributed under the GNU LGPL license. 
	//
	//  Modified:
	//
	//    16 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by JC Young, Christoph Minder.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    MA Porter, DJ Winstanley,
	//    Remark AS R30:
	//    A Remark on Algorithm AS76:
	//    An Integral Useful in Calculating Noncentral T and Bivariate
	//    Normal Probabilities,
	//    Applied Statistics,
	//    Volume 28, Number 1, 1979, page 113.
	//
	//    JC Young, Christoph Minder,
	//    Algorithm AS 76: 
	//    An Algorithm Useful in Calculating Non-Central T and 
	//    Bivariate Normal Distributions,
	//    Applied Statistics,
	//    Volume 23, Number 3, 1974, pages 455-457.
	//
	//  Parameters:
	//
	//    Input, double X, FX, the parameters of the function.
	//
	//    Output, double TFN, the value of the T-function.
	//
	{
		const NG = 5;

		let fxs, i, r1, r2, rt, value, x1, x2, xs;
		let r = [
			0.1477621, 
			0.1346334, 
			0.1095432, 
			0.0747257, 
			0.0333357 ];
		let tp = 0.159155;
		let tv1 = 1.0E-35;
		let tv2 = 15.0;
		let tv3 = 15.0;
		let tv4 = 1.0E-05;
		let u = [
			0.0744372, 
			0.2166977, 
			0.3397048,
			0.4325317, 
			0.4869533 ];


	//
	//  Test for X near zero.
	//
		if ( Math.abs ( x ) < tv1 )
		{
			value = tp * Math.atan ( fx );
			return value;
		}
	//
	//  Test for large values of abs(X).
	//
		if ( tv2 < Math.abs ( x ) )
		{
			value = 0.0;
			return value;
		}
	//
	//  Test for FX near zero.
	//
		if ( Math.abs ( fx ) < tv1 )
		{
			value = 0;
			return value;
		}
	//
	//  Test whether abs ( FX ) is so large that it must be truncated.
	//
		xs = - 0.5 * x * x;
		x2 = fx;
		fxs = fx * fx;
	//
	//  Computation of truncation point by Newton iteration.
	//
		if ( tv3 <= Math.log ( 1 + fxs ) - xs * fxs )
		{
			x1 = 0.5 * fx;
			fxs = 0.25 * fxs;

			for ( ; ; )
			{
				rt = fxs + 1;

				x2 = x1 + ( xs * fxs + tv3 - Math.log ( rt ) ) 
				/ ( 2 * x1 * ( 1 / rt - xs ) );

				fxs = x2 * x2;

				if ( Math.abs ( x2 - x1 ) < tv4 )  {
					break;
				}
				x1 = x2;
			}
		}
	//
	//  Gaussian quadrature.
	//
		rt = 0.0;
		for ( i = 0; i < NG; i++ )
		{
			r1 = 1 + fxs * Math.pow ( 0.5 + u[i], 2 );
			r2 = 1 + fxs * Math.pow ( 0.5 - u[i], 2 );

			rt = rt + r[i] * ( Math.exp ( xs * r1 ) / r1 + Math.exp ( xs * r2 ) / r2 );
		}

		value = rt * x2 * tp;

		return value;

	}
	//****************************************************************************80
	
} 
